using System;
using System.Reflection;
using System.CodeDom.Compiler;

using Microsoft.CSharp;

using Seva.ApplicationLogs; 
using Seva.Data;

namespace Seva.Math.DistributedIntegration
{
    [Serializable]
    public class IntegrationRequest//: MarshalByRefObject//: ISerializable
    {
        public String functionString;
        public double x1, x2, eps;   
        
        public IntegrationRequest()
        {
        }           
            
        public IntegrationRequest(String afunctionString, double ax1, double ax2, double aeps)     
        {
            functionString = afunctionString;
            x1 = ax1;
            x2 = ax2;
            eps =aeps;
        }
        
        public override String ToString()
        {
            return String.Format("f(x)={0}, {1:0.0000}<=x<={2:0.0000}, eps={3}", 
                                this.functionString, this.x1, this.x2, this.eps);
        }   
    };

    [Serializable]
    public class IntegrationAnswer
    {
        public IntegrationRequest request;
        public double result;   
        
        public int steps;
            
        public IntegrationAnswer()
        {
        }
        
        public IntegrationAnswer(double aresult, int asteps, IntegrationRequest arequest)
        {
            result  = aresult;
            steps = asteps;
            request = arequest;
        }
        
        public override String ToString()
        {
            return String.Format("F({0}) = {1:0.0000}  [n = {2}]", 
                                this.request.ToString(), this.result, this.steps);
        }
    };      
    
    public interface IMathFunction
    {
        double f(double x);
    }
    
    class CompiledFunction : MarshalByRefObject
    {
        readonly IMathFunction function;
        
        public CompiledFunction(string functionCode)//, string [] assemblies)
        {   
            //if (code == null || assemblies == null || speakOut == null )
                //throw new ArgumentNullException();
            
            string className = "CompilerTest.Test";
            string [] assemblies = {"system.dll"};          
            
            string code = 
@"
using System;
using Seva.Math.DistributedIntegration;         
            
namespace CompilerTest
{
      public class Test : IMathFunction
      {
        const double pi = Math.PI;  
            
        public double sin(double x) {return Math.Sin(x);}
        public double cos(double x) {return Math.Cos(x);}
        public double sqrt(double x) {return Math.Sqrt(x);}
        public double abs(double x) {return Math.Abs(x);}
        public double exp(double x) {return Math.Exp(x);}
        public double pow(double x, double y) {return Math.Pow(x, y);}
        public double ln(double x) {return Math.Log(x);}
        public double log(double x, double y) {return Math.Log(x, y);}
            
          public double f(double x)
          {
            return (" + functionCode + @");
          }
      }
}";
        
            CSharpCodeProvider codeCompiler = new CSharpCodeProvider();
            CompilerParameters parameters = new CompilerParameters(assemblies);
            string path = Assembly.GetExecutingAssembly().Location;
            parameters.ReferencedAssemblies.Add(path);
            CompilerResults results = codeCompiler.CompileAssemblyFromSource(parameters, code);
            if( results.Errors.HasErrors )
            {   
                foreach( CompilerError error in results.Errors )
                    Log.Message(String.Format("Line:{0:d},Error:{1}\n", error.Line, error.ErrorText));
                Log.Error("Error in compiled function.");
            }   

            object objFunction = results.CompiledAssembly.CreateInstance(className);
            if (objFunction == null)
                Log.Error("Error when creating " + className);
            function = objFunction as IMathFunction;
            if (function == null)
                Log.Error("IFunction does not implemented.");

        }
        
        public double f(double x)
        {
            //Assert(function!=null);
            return function.f(x);
        }
    }       
    
    class Integration
    {
        private Cache<String, CompiledFunction> functions;
        
        public Integration()
        {
            functions = new Cache<String, CompiledFunction>(100);
        }           
        
        public double Sum(CompiledFunction function, double x1, double x2, int n)
        {
            double result = 0;
            double step = (x2-x1)/n;
            double x;
            int i;
            for(i=0, x=x1+step/2; i<n; i++, x+=step)
                result += function.f(x) * step;
            return result;
        }
        
        public double Calc(CompiledFunction function, double x1, double x2, double eps, ref int steps)
        {
            int n = 1;
            double last = 0;
            double cur = Sum(function, x1, x2, n);
            do
            {
                last = cur;
                n *= 2;
                cur = Sum(function, x1, x2, n);
            }
            while (System.Math.Abs(last-cur)>eps);          
            steps = n;
            return cur;         
            
        }           
        
        public IntegrationAnswer Run(IntegrationRequest request)
        {
            String functionCode  = request.functionString.Replace(" ", "");
            request.functionString = functionCode;
            //
            CompiledFunction function = functions.Find(functionCode);
            //
            if (function == null)                
                function = functions.Add(functionCode, new CompiledFunction(functionCode));
           // else      Log.Message("hits!"); 
            int steps = 0;
            double result  = Calc(function, request.x1, request.x2, request.eps, ref steps);//.f(request.x1);
            //
            IntegrationAnswer answer = new IntegrationAnswer(result, steps, request);
            return answer;
        }           
    }       
}       