using System;
using System.Messaging;
using System.Collections.Generic;

namespace Seva.Msmq
{
    //        
    public delegate AnswerType ProcessRequestEventHandler
        <RequestType, AnswerType>(Object sender, RequestType request, 
        MessageQueue queueResponse);

    //        
    public delegate void ProcessAnswerEventHandler<RequestType, AnswerType>
        (Object sender, RequestType request, AnswerType answer);    

    public class MsmqUser
    {
        //   
        private bool recoverable = false;            
        public bool Recoverable 
        {
            get 
            {
                return recoverable;
            }
            set 
            {
                recoverable = value;
            } 
        }
    }

    //   ,   MSMQ   
    //    
    class MsmqClient<RequestType, AnswerType> : MsmqUser
    {
        //       
        private MessageQueue queueSend;
        private MessageQueue queueReceive;

        //      
        private XmlMessageFormatter requestFormatter;
        private XmlMessageFormatter answerFormatter;

        //   
        private Dictionary<String, RequestType> messages;  

        public Dictionary<String, RequestType> Messages 
        {
            get { return messages;}
        }

        // ,    
        public event ProcessAnswerEventHandler<RequestType, AnswerType> 
            ProcessAnswer;

        // ,        
        public MsmqClient(String queueSendName, String queueReceiveName)
        {
            requestFormatter =  new XmlMessageFormatter(); 
            answerFormatter =  new XmlMessageFormatter(
                new Type[]{typeof(AnswerType)});
            //                                 
            messages = new Dictionary<String,RequestType>();
            //     ,    
            queueSend = MsmqTools.CreateQueue(queueSendName,
                MsmqTools.TransactionalQueue);
                
            //     ,   
            if (queueReceiveName.Length!=0)
            {
                queueReceive = MsmqTools.CreateQueue(queueReceiveName);         
                queueReceive.Formatter = answerFormatter;
                queueReceive.MessageReadPropertyFilter.CorrelationId = true;
            }
        }   
  
        //    
        public void BeginReceive()
        {
            //    ,    
            //               
            queueReceive.PeekCompleted += OnPeek;
            //      
            queueReceive.BeginPeek();          
        }   

         //     
        public void EndReceive()
        {
            //   
            queueReceive.PeekCompleted -= OnPeek;
        }   

        //    
        public void Send(RequestType request)
        {            
            //   
            Message message = new Message(request, requestFormatter);             
            message.Label = "Request";            

            //        ,
            //    ResponseQueue 
            message.ResponseQueue = queueReceive;
            
            message.Recoverable = Recoverable;
            //  ;     
            //  ,  - 
            //  MessageQueueTransactionType.Single
            queueSend.Send(message, MessageQueueTransactionType.Single);
            
            //  message.Id    
            //       
            //    
            messages.Add(message.Id, request);          
        }   

        //    PeekComplete
        public void OnPeek(Object source, PeekCompletedEventArgs asyncResult)
        {
            //    MSMQ 
            MessageQueueTransaction transaction = new MessageQueueTransaction();
            
            //  
            transaction.Begin();      

            try
            {                   
                queueReceive.EndPeek(asyncResult.AsyncResult);

                //       
                Message message = queueReceive.Receive(transaction);

                //   CorrelationId    
                //   
                String messageId = message.CorrelationId;

                if (messages.ContainsKey(messageId))

                {                       
                    if (message.Body is AnswerType)
                    {
                        //     
                        //                               
                        AnswerType answer = (AnswerType) message.Body;
                        ProcessAnswer(this, messages[messageId], answer);
                    };                  
                    messages.Remove(messageId);
                }                       
                //   
                BeginReceive();
                
                //   
                transaction.Commit();
            }
            catch (Exception e)
            {
                //   
                transaction.Abort();
                throw e;
            }               
        }   
    }          

    //   ,   MSMQ   
    //    ,
    public class MsmqServer<RequestType, AnswerType>: MsmqUser
    {
        //   
        private MessageQueue queueReceive;

        private XmlMessageFormatter requestFormatter;
        private XmlMessageFormatter answerFormatter;

        // ,           
        public event ProcessRequestEventHandler<RequestType, AnswerType>
            ProcessMessage;                

        //     
        public MsmqServer(String queueReceiveName)
        {
            requestFormatter =  new XmlMessageFormatter(
                new Type[]{typeof(RequestType)});
            answerFormatter =  new XmlMessageFormatter(); 
            
            //    ,    
            queueReceive = MsmqTools.CreateQueue(queueReceiveName,
                MsmqTools.TransactionalQueue);         
            queueReceive.Formatter = requestFormatter;
        }   
        
        //     
        public void BeginReceive()
        {
            queueReceive.PeekCompleted += OnPeek;
            queueReceive.BeginPeek();           
        }   

        public void EndReceive()
        {            
            queueReceive.PeekCompleted -= OnPeek;
        }   

        //   PeekCompleted   
        public void OnPeek(Object source, PeekCompletedEventArgs asyncResult)
        {
            //    MSMQ 
            MessageQueueTransaction transaction = new MessageQueueTransaction(); 
            
            //         
            transaction.Begin();

            try
            {               
                queueReceive.EndPeek(asyncResult.AsyncResult);

                //  c    
                Console.WriteLine("receive:");
                Message message = queueReceive.Receive(transaction);
                //MessageQueueTransactionType.Single
                Console.WriteLine("received");
                //   ResponseQueue    , 
                //      
                MessageQueue queueResponse = message.ResponseQueue;

                if (message.Body is RequestType)
                {
                    RequestType request = (RequestType) message.Body;
                    //     
                    AnswerType answer = ProcessMessage(this, request, 
                        queueResponse);

                    if (queueResponse!=null && answer!=null)
                    {               
                        Message answerMessage = new Message(answer, 
                            answerFormatter); 
                        answerMessage.Label = "Answer";
                        answerMessage.CorrelationId = message.Id;
                        answerMessage.Recoverable = Recoverable;

                        //     
                        queueResponse.Send(answerMessage, transaction);                   
                    }   
                }

                //   
                BeginReceive();
                //   
                transaction.Commit();
            }
            catch (Exception e)
            {
                //      
                Console.WriteLine(e);
                //transaction.Abort();  
                throw e;        
            }
        }   
    }       

    //     

    public abstract class MsmqTools
    {
        public const bool TransactionalQueue = true;
        public const bool NonTransactionalQueue = false;

        
        static public MessageQueue CreateQueue(String queueName)
        {
            return CreateQueue(queueName, TransactionalQueue);
        }

        //         
        static public MessageQueue CreateQueue(String queueName, 
            bool transactional)
        {
            MessageQueue messageQueue;
            
            //      ,
            //        
            try 
            {
                if (!MessageQueue.Exists(queueName))
                {
                    MessageQueue.Create(queueName, transactional);
                }
            }    
            catch(Exception)
            {
            }            
            
            messageQueue = new MessageQueue(queueName);
            
            return messageQueue;
        }
    }   
}