using System;
using System.IO;
using System.Reflection;
using System.Text;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;
using System.Collections.Generic;
using System.Security.Cryptography;
using System.Security;

using Microsoft.Web.Services3;
using Microsoft.Web.Services3.Design;
using Microsoft.Web.Services3.Security;
using Microsoft.Web.Services3.Security.Tokens;

[assembly:AssemblyVersionAttribute("1.0.0.0")]

namespace Seva.WS.Users
{
    public class UsersListManager: UsernameTokenManager 
    {
        private UsersList users;
            
        public UsersListManager()
        {
            users = new UsersList();
        }
    
        public UsersListManager(XmlNodeList configData): base(configData)
        {            
            string fileName = configData[0].Attributes["file"].Value;            
            users = UsersList.Load(fileName);
        }
         
        protected override string AuthenticateToken(UsernameToken token)
        {            
            if (!users.Users.ContainsKey(token.Username))
                return null;
            return users.Users[token.Username];
        }
        
        protected override void VerifyPlainTextPassword(UsernameToken token, 
            string authenticatedPassword) 
        {        
            if (token==null)
                throw new ArgumentNullException("token is null"); 
 
            String hashed = Utils.HashedPassword(token.Password);
                        
            if (authenticatedPassword==null || authenticatedPassword=="" || 
                hashed!=authenticatedPassword)
                throw new Exception("Passwords does not match ");             
        }
    }

    [XmlRoot("user")]
    public struct UserCredential
    {
        [XmlAttribute("username")]    
        public string Username;
            
        [XmlAttribute("password")]    
        public string Password;
    }
    
    public class UsernameClientAssertion: SecurityPolicyAssertion
    {        
        private UserCredential credential;
            
        public UsernameClientAssertion()
        {
        }
        
        public override void ReadXml(XmlReader reader, 
            IDictionary<string, Type> extensions)
        {                        
            string fileName = reader.GetAttribute("file");
            if (fileName == null)
                throw new Exception("Attribute 'file' not found in policy");
            
            //   -       
            reader.Read();
                     
            using (StreamReader timeReader = new StreamReader(fileName))
            {
                XmlSerializer serializer = new XmlSerializer(typeof(UserCredential));
                using (XmlTextReader xmlReader = new XmlTextReader(timeReader))
                {
                    credential = (UserCredential) serializer.Deserialize(xmlReader);
                }    
            }
        }
        
        public override SoapFilter 
           CreateClientOutputFilter(FilterCreationContext context)
        {
            return new UsernameClientInputFilter(this, credential, context);
        }

        public override SoapFilter 
               CreateClientInputFilter(FilterCreationContext context)
        {            
            return null;
        }

        public override SoapFilter 
               CreateServiceInputFilter(FilterCreationContext context)
        {
            return null;
        }

        public override SoapFilter 
               CreateServiceOutputFilter(FilterCreationContext context)
        {
            return null;
        }
    }
    
    class UsernameClientInputFilter : SendSecurityFilter
    {
        private UserCredential credential;
        
        public UsernameClientInputFilter(UsernameClientAssertion assertion, 
            UserCredential credential, FilterCreationContext filterContext)
            : base(assertion.ServiceActor, true, assertion.ClientActor)
        {
            this.credential = credential;
        }

        public override void SecureMessage(SoapEnvelope envelope, Security security)
        {
            UsernameToken userToken = new UsernameToken(credential.Username, 
                credential.Password, PasswordOption.SendHashed);
            
            security.Tokens.Add(userToken);            
        }
    }
 
    
    
    [XmlRoot("users")]    
    public class UsersList: IXmlSerializable
    {
        private Dictionary<string, string> usersField;           
        
        public Dictionary<string, string> Users
        {
            get {return usersField;}
        }
        
        public UsersList()
        {
            usersField = new Dictionary<string, string>();
        }
        
        public void ReadXml(XmlReader reader)
        {       
            reader.Read();                            
            while (reader.NodeType != XmlNodeType.EndElement)
            {   
                Users.Add(reader.GetAttribute("username"), reader.GetAttribute("password"));
                reader.Read();                
            }
        }
        
        public void WriteXml(XmlWriter writer)
        {    
            foreach  (string user in Users.Keys)
            {
                writer.WriteStartElement("user");
                writer.WriteAttributeString("username", user);
                writer.WriteAttributeString("password", this.Users[user]);
                writer.WriteEndElement();
            }
        }
        
        public XmlSchema GetSchema()
        {
            return null;
        }
        
        public void NewUser(string userName, string password, bool hashed)
        {           
            if (Users.ContainsKey(userName))
            {
                throw new ApplicationException("Duplicate username.");
            }
            
            if (hashed)
            {
                password = Utils.HashedPassword(password);
            }
            
            Users.Add(userName, password);
        }       
        
        public static UsersList Load(string fileName)
        {
            XmlSerializer serializer = new XmlSerializer(typeof(UsersList));
            
            using (StreamReader reader = new StreamReader(fileName))
            {
                return (UsersList) serializer.Deserialize(reader);
            };
        }    
            
        public void Save(string fileName)    
        {
            XmlSerializer serializer = new XmlSerializer(typeof(UsersList));
            
            using (StreamWriter writer = new StreamWriter(fileName))
            {
                serializer.Serialize(writer, this);
            };
        }        
    }    
    
    static class Utils
    {
        public static string HashedPassword(string password)
        {
            SHA1CryptoServiceProvider crypto = new SHA1CryptoServiceProvider();

            byte[] inputBuffer =  Encoding.Unicode.GetBytes(password);        
            return Convert.ToBase64String(crypto.ComputeHash(inputBuffer));
        }
    }
}    