//  SevaRemotingEncrypted.cs
using System;
using System.IO;
using System.Collections;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Tcp;
using System.Runtime.Remoting.Messaging; 

using Seva.Security.Encryption;

namespace Seva.Remoting.Encryption
{       
    class EncryptedClientChannelSinkProvider: IClientChannelSinkProvider
    {
        private IClientChannelSinkProvider next;
            
        public IClientChannelSinkProvider Next 
        { 
            get { return next; } 
            set { next = value; }
        }
       
        private SymmetricEncryptor encryptor;
            
        public EncryptedClientChannelSinkProvider(IDictionary properties, ICollection providerData)
        {
            string keyFile = (string) properties["key"];
            Console.WriteLine("Client key: [{0}]", keyFile);
            encryptor = new SymmetricEncryptor(keyFile);
        }
        
        public IClientChannelSink CreateSink(IChannelSender channel, 
            string url, object remoteChannelData) 
        {
            IClientChannelSink next = Next.CreateSink(channel,
                url, remoteChannelData);
            return new EncryptedClientChannelSink(encryptor, next);    
        }
    }
    
    class EncryptedServerChannelSinkProvider : IServerChannelSinkProvider
    {
        private IServerChannelSinkProvider next;
        private SymmetricEncryptor encryptor;
            
        public IServerChannelSinkProvider Next 
        { 
            get { return next; } 
            set { next = value; }
        }    
        
        public EncryptedServerChannelSinkProvider(IDictionary properties, ICollection providerData)
        {
            string keyFile = (string) properties["key"];
            Console.WriteLine("Server key: [{0}]", keyFile);
            encryptor = new SymmetricEncryptor(keyFile);
        }
        
        //   
        public IServerChannelSink CreateSink(IChannelReceiver channel)
        {            
           // IServerChannelSinkProvider provider = new BinaryServerFormatterSinkProvider();        
            //Console.WriteLine(Next);
            //IServerChannelSink nextSink = 
                //ChannelServices.CreateServerChannelSinkChain(Next, channel);
            IServerChannelSink nextSink = Next.CreateSink(channel);            
            return new EncryptedServerChannelSink(channel, encryptor, nextSink);
        }           
        
        //   
        public void GetChannelData(IChannelDataStore channelData)
        {
        }        
    }
    
    class EncryptedServerChannelSink : IServerChannelSink
    {
        public IDictionary Properties
        {   
            get { return null; }
        }
        
        private IServerChannelSink nextSink;
        public IServerChannelSink NextChannelSink
        { 
            get { return nextSink;  }
        }
        
        private SymmetricEncryptor encryptor;
               
        public EncryptedServerChannelSink(IChannelReceiver channel, 
            SymmetricEncryptor channelEncryptor, IServerChannelSink next)
        {
            encryptor = channelEncryptor;
            nextSink = next;
        }

        //    
        public ServerProcessing ProcessMessage(IServerChannelSinkStack sinkStack,
            IMessage requestMsg, ITransportHeaders requestHeaders,
            Stream requestStream, out IMessage responseMsg, 
            out ITransportHeaders responseHeaders, out Stream responseStream)
        {               
            Stream plainStream = null;
            MemoryStream decodedStream = new MemoryStream();
            
            requestStream = Utils.ReadAllStream(requestStream);
            
            //   
            encryptor.Decrypt(requestStream, decodedStream);
            decodedStream.Position = 0;
                        
            ServerProcessing result = nextSink.ProcessMessage(sinkStack,
                requestMsg, requestHeaders, decodedStream, out responseMsg, 
                out responseHeaders, out plainStream);
            
            //   
            responseStream = new MemoryStream();
            encryptor.Encrypt(plainStream, responseStream);            
            responseStream.Position = 0;
            //
            return result;
        }        
        
        public void AsyncProcessResponse(IServerResponseChannelSinkStack sinkStack,
            Object state, IMessage msg, ITransportHeaders headers, Stream stream)
        {
            //      
            //   
            throw new NotSupportedException();
        }
        
        public Stream GetResponseStream(IServerResponseChannelSinkStack sinkStack,
            Object state, IMessage msg, ITransportHeaders headers)
        { 
            return nextSink.GetResponseStream(sinkStack, state, msg, headers); 
        }
         
    } // EncryptedServerChannelSink 
    
   
    class EncryptedClientChannelSink: IClientChannelSink
    {
        public IDictionary Properties
        {   
            get { return null; }
        }
        
        private IClientChannelSink nextSink;
        
        public IClientChannelSink NextChannelSink
        { 
            get { return nextSink;  }
        }
               
        private SymmetricEncryptor encryptor;
               
        public EncryptedClientChannelSink(SymmetricEncryptor channelEncryptor,
            IClientChannelSink next)
        {
            encryptor = channelEncryptor; 
            nextSink = next;
        }            
        
        public void ProcessMessage(IMessage message, 
            ITransportHeaders requestHeaders, Stream requestStream,
            out ITransportHeaders responseHeaders, out Stream responseStream)
        {                        
            MemoryStream encryptedStream = new MemoryStream();            
            Stream serverStream = null;
            
            //   
            encryptor.Encrypt(requestStream, encryptedStream);
            encryptedStream.Position = 0;
            
            nextSink.ProcessMessage(message, requestHeaders, encryptedStream, 
                out responseHeaders, out serverStream);
            
            serverStream = Utils.ReadAllStream(serverStream);
            
            //   
            responseStream = new MemoryStream();
            encryptor.Decrypt(serverStream, responseStream);
            responseStream.Position = 0;
        }  
        
        public Stream GetRequestStream(IMessage msg, ITransportHeaders headers)
        { 
            return null; 
        }               
        
        public void AsyncProcessRequest(IClientChannelSinkStack sinkStack, 
            IMessage message, ITransportHeaders headers, Stream stream) 
        {
            throw new NotSupportedException();
        }        
        
        public void AsyncProcessResponse(
            IClientResponseChannelSinkStack sinkStack, object state, 
            ITransportHeaders headers, Stream stream)
        {
            throw new NotSupportedException();
        }        
    } // EncryptedClientChannelSink
    
    abstract class Utils
    {
        //  ,      
        //     . 
        //     , 
        //    Length
        public static MemoryStream ReadAllStream(Stream input)
        {
            byte[] buffer = new byte [1024];
            MemoryStream output = new MemoryStream();
            while (true)
            {
                int read = input.Read(buffer, 0, buffer.Length);
                output.Write(buffer, 0, read);                    
                if (read < buffer.Length) break;
                
            }    
            output.Position = 0;            
            return output;
        }
    }
}
// SevaRemotingEncrypted.cs