using System;
using System.Messaging;
using System.Collections.Generic;

namespace Seva.Msmq
{
    enum QueueType  {NonTransactional, Transactional};
    enum QueueFormatter {Binary, Xml};
        
    //        
    delegate AnswerType ProcessRequestEventHandler
        <RequestType, AnswerType>(Object sender, RequestType request, 
        MessageQueue queueResponse);

    //        
    delegate void ProcessAnswerEventHandler<RequestType, AnswerType>
        (Object sender, RequestType request, AnswerType answer);    

    class MsmqUser<RequestType, AnswerType>
    {
        //   
        private bool recoverable = false;            
        public bool Recoverable 
        {
            get 
            {
                return recoverable;
            }
            set 
            {
                recoverable = value;
            } 
        }

        //      
        protected IMessageFormatter requestFormatter;
        protected IMessageFormatter answerFormatter;
        //
        public MsmqUser(QueueFormatter formatterType)
        {
            if (formatterType == QueueFormatter.Xml)
            {
                requestFormatter =  new XmlMessageFormatter(
                    new Type[]{typeof(RequestType)}); 
                answerFormatter =  new XmlMessageFormatter(
                    new Type[]{typeof(AnswerType)});
            }        
            else
            {
                requestFormatter =  new BinaryMessageFormatter(); 
                answerFormatter =  new BinaryMessageFormatter();
            }
        }
    }

    //   ,   MSMQ   
    //    
    class MsmqClient<RequestType, AnswerType> : 
        MsmqUser<RequestType, AnswerType>, IDisposable
    {
        //       
        private MessageQueue queueSend;
        private MessageQueue queueReceive;        

        //   
        private Dictionary<String, RequestType> messages;  

        public Dictionary<String, RequestType> Messages 
        {
            get { return messages;}
        }

        // ,    
        public event ProcessAnswerEventHandler<RequestType, AnswerType> 
            ProcessAnswer;

        // ,        
        public MsmqClient(String queueSendName, String queueReceiveName, 
            QueueFormatter formatterType): base(formatterType)
        {
            //                                     
            messages = new Dictionary<String,RequestType>();
            //     ,    
            queueSend = MsmqTools.CreateQueue(queueSendName,
                QueueType.Transactional);
                
            //     ,   
            if (queueReceiveName != null)
            {
                queueReceive = MsmqTools.CreateQueue(queueReceiveName);         
                queueReceive.Formatter = answerFormatter;
                //     CorrelationId
                queueReceive.MessageReadPropertyFilter.CorrelationId = true;
            }
            else
            {
                queueReceive = null;
            }
        }   
        
        //   
        public void Dispose()
        {
            queueSend.Close();
            queueSend.Dispose();
            
            if (queueReceive != null)
            {
                queueReceive.Close();
                queueReceive.Dispose();
            }
        }

  
        //    
        public void BeginReceive()
        {
            //    ,    
            //               
            queueReceive.PeekCompleted += OnPeek;
            //      
            queueReceive.BeginPeek();          
        }   

         //     
        public void EndReceive()
        {
            //   
            queueReceive.PeekCompleted -= OnPeek;            
        }   

        public void Send(RequestType request)
        {            
            Send(request, "");
        }
        //    
        public void Send(RequestType request, string label)
        {            
            //   
            Message message = new Message(request, requestFormatter);             
            message.Label = label;            

            //        ,
            //    ResponseQueue 
            message.ResponseQueue = queueReceive;
            
            message.Recoverable = Recoverable;
            //  ;     
            //  ,  - 
            //  MessageQueueTransactionType.Single
            queueSend.Send(message, MessageQueueTransactionType.Single);
            
            //  message.Id    
            //       
            //    
            if (queueReceive != null)
                messages.Add(message.Id, request);          
        }   

        //    PeekComplete
        public void OnPeek(Object source, PeekCompletedEventArgs asyncResult)
        {
            //    MSMQ 
            MessageQueueTransaction transaction = new MessageQueueTransaction();
            
            //  
            transaction.Begin();      

            try
            {                   
                queueReceive.EndPeek(asyncResult.AsyncResult);

                //       
                Message message = queueReceive.Receive(transaction);

                //   CorrelationId    
                //   
                String messageId = message.CorrelationId;

                //       
                if (messages.ContainsKey(messageId))
                {                       
                    if (message.Body is AnswerType)
                    {
                        //     
                        //                               
                        AnswerType answer = (AnswerType) message.Body;
                        ProcessAnswer(this, messages[messageId], answer);
                    };                
                    
                    messages.Remove(messageId);
                }                       
                //   
                BeginReceive();
                
                //   
                transaction.Commit();
            }
            catch (Exception e)
            {
                //   
                transaction.Abort();
                throw e;
            }               
        }   
    }          

    //   ,   MSMQ   
    //    ,
    class MsmqServer<RequestType, AnswerType>: 
        MsmqUser<RequestType, AnswerType>, IDisposable
    {
        //   
        private MessageQueue queueReceive;

        // ,           
        public event ProcessRequestEventHandler<RequestType, AnswerType>
            ProcessMessage;                

        //     
        public MsmqServer(String queueReceiveName, QueueFormatter formatterType): 
            base(formatterType)
        {                       
            //    ,    
            queueReceive = MsmqTools.CreateQueue(queueReceiveName,
                QueueType.Transactional);         
            queueReceive.Formatter = requestFormatter;
        }

        //   
        public void Dispose()
        {
            queueReceive.Close();
            queueReceive.Dispose();
        }        
        
        //     
        public void BeginReceive()
        {
            queueReceive.PeekCompleted += OnPeek;
            queueReceive.BeginPeek();           
        }   

        public void EndReceive()
        {            
            queueReceive.PeekCompleted -= OnPeek;
        }   

        //   PeekCompleted   
        public void OnPeek(Object source, PeekCompletedEventArgs asyncResult)
        {
            //    MSMQ 
            MessageQueueTransaction transaction = new MessageQueueTransaction();             
            
            //         
            transaction.Begin();

            try
            {               
                queueReceive.EndPeek(asyncResult.AsyncResult);

                //  c    
                Message message = queueReceive.Receive(transaction);
                
                //   ResponseQueue    , 
                //      
                MessageQueue queueResponse = message.ResponseQueue;
                try
                {
                    if (message.Body is RequestType)
                    {
                        RequestType request = (RequestType) message.Body;
                        //     
                        AnswerType answer = ProcessMessage(this, request, 
                            queueResponse);
                        if ((queueResponse != null) && (answer != null))
                        {               
                            Message answerMessage = new Message(answer, 
                                answerFormatter); 
                            answerMessage.Label = "Answer";
                            answerMessage.CorrelationId = message.Id;
                            answerMessage.Recoverable = Recoverable;

                            //     
                            queueResponse.Send(answerMessage, transaction);                   
                        }   
                    }
                }                    
                finally
                {
                    if (queueResponse != null)
                    {
                        queueResponse.Close();
                        queueResponse.Dispose();
                    }
                };                    

                //   
                BeginReceive();
                //   
                transaction.Commit();
            }
            catch (Exception e)
            {
                //     
                Console.WriteLine(e);
                transaction.Abort();  
                throw e;        
            }
        }   
    }       

    //     
    static class MsmqTools
    {             
        static public MessageQueue CreateQueue(String queueName)
        {
            return CreateQueue(queueName, QueueType.Transactional);
        }

        //         
        static public MessageQueue CreateQueue(String queueName, 
            QueueType type)
        {
            MessageQueue messageQueue;
            
            //      ,
            //        
            try 
            {
                if (!MessageQueue.Exists(queueName))
                {
                    MessageQueue.Create(queueName, type == QueueType.Transactional);
                }
            }    
            catch(Exception)
            {
            }            
            
            messageQueue = new MessageQueue(queueName);
            
            return messageQueue;
        }
    }   
}